// hzip example illustrating compression of structured grids

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/time.h>
#include <vector>
#include "hzip.h"

using namespace std;

#if !defined TYPE || !defined HZTYPE
  #error "must define TYPE and HZTYPE"
#endif

#ifndef WITHOUT_ZLIB
  #define CODEC HZN_CODEC_ZLIB
#else
  #define CODEC HZN_CODEC_BASE
#endif

static double
now()
{
  struct timeval tv;
  gettimeofday(&tv, 0);
  return tv.tv_sec + 1e-6 * tv.tv_usec;
}

static bool
read(FILE* file, vector<TYPE>& grid, unsigned dims, const unsigned count[])
{
  int length = 1;
  for (unsigned i = 0; i < dims; i++)
    length *= count[i];
  grid = vector<TYPE>(length);
  if ((int)fread(&grid[0], sizeof(grid[0]), length, file) != length) {
    fprintf(stderr, "error reading file\n");
    return false;
  }
  return true;
}

static int
usage(char* argv[])
{
  fprintf(stderr, "Usage: %s <infile> <n> <dim1 ... dimn> [outfile]\n", *argv);
  return EXIT_FAILURE;
}

int
main(int argc, char* argv[])
{
  if (argc < 4)
    return usage(argv);

  char* infile = argv[1];
  unsigned dims;
  if (sscanf(argv[2], "%u", &dims) != 1 || argc < (int)dims + 3)
    return usage(argv);

  unsigned nodes = 1;
  unsigned cells = 1;
  unsigned count[HZ_DIMS_MAX];
  for (unsigned i = 0; i < dims; nodes *= count[i], cells *= count[i] - 1, i++)
    if (sscanf(argv[i + 3], "%u", &count[i]) != 1)
      return usage(argv);

  char* outfile;
  switch (argc - (dims + 3)) {
    case 0:
      outfile = 0;
      break;
    case 1:
      outfile = argv[argc - 1];
      break;
    default:
      return usage(argv);
  }

  // read uncompressed quad or hex mesh
  fprintf(stderr, "reading\n");
  FILE* file = fopen(infile, "rb");
  if (!file) {
    fprintf(stderr, "cannot open file\n");
    return EXIT_FAILURE;
  }
  vector<TYPE> grid;
  if (!read(file, grid, dims, count))
    return EXIT_FAILURE;
  fclose(file);
  fprintf(stderr, "bytes=%u\n", (unsigned)(grid.capacity() * sizeof(grid[0])));

  // construct mesh connectivity
  vector<int> mesh(cells << dims);
  unsigned perm = hzip_mesh_construct(&mesh[0], dims, count, 0);

  HZNstream* stream;
  unsigned char* buffer;
  int inbytes = nodes * sizeof(grid[0]);
  int outbytes;
  double t;

  t = now();
  if (outfile) {
    // compress grid to file
    fprintf(stderr, "compressing grid to file\n");
    file = fopen(outfile, "wb");
    if (!file) {
      fprintf(stderr, "cannot create file\n");
      return EXIT_FAILURE;
    }
    buffer = 0;
  }
  else {
    // compress grid to memory
    fprintf(stderr, "compressing grid to memory\n");
    buffer = new unsigned char[inbytes];
    file = 0;
  }
  // execute compression
  if (file)
    stream = hzip_node_create_file(file, perm, 0, dims, nodes, HZTYPE, CODEC, 0);
  else
    stream = hzip_node_create_mem(buffer, inbytes, perm, 0, dims, nodes, HZTYPE, CODEC, 0);
  if (!stream ||
      hzip_node_write(stream, &grid[0], &mesh[0], cells) < 0 ||
      (outbytes = hzip_node_close(stream)) < 0) {
    fprintf(stderr, "compression failed\n");
    return EXIT_FAILURE;
  }
  t = now() - t;
  fprintf(stderr, "in=%u out=%u ratio=%.2f bpn=%.2f seconds=%.3f nps=%.0f\n", inbytes, outbytes, (double)inbytes / outbytes, outbytes * 8.0 / nodes, t, nodes / t);

  if (file)
    fclose(file);

#ifndef WITHOUT_DECOMPRESSION
  t = now();
  if (outfile) {
    // decompress grid from file
    fprintf(stderr, "decompressing grid from file\n");
    file = fopen(outfile, "rb");
    if (!file) {
      fprintf(stderr, "cannot open file\n");
      return EXIT_FAILURE;
    }
  }
  else {
    // decompress grid from memory
    fprintf(stderr, "decompressing grid from memory\n");
  }
  // ececute decompression
  vector<TYPE> copy(nodes);
  if (file)
    stream = hzip_node_open_file(file, perm, 0);
  else
    stream = hzip_node_open_mem(buffer, inbytes, perm, 0);
  int status;
  if (!stream ||
       hzip_node_dimensions(stream) != dims ||
       hzip_node_count(stream) != nodes ||
       hzip_node_type(stream) != HZTYPE ||
       hzip_node_read(stream, &copy[0], &mesh[0], cells) < 0 ||
       (status = hzip_node_close(stream)) < 0) {
    fprintf(stderr, "decompression failed\n");
    return EXIT_FAILURE;
  }
  if (status != outbytes)
    fprintf(stderr, "ERROR: unexpected byte count\n");
  t = now() - t;
  fprintf(stderr, "seconds=%.3f nps=%.0f\n", t, nodes / t);

  if (file)
    fclose(file);

  // validate data
  fprintf(stderr, "validating\n");
  t = now();
  if (memcmp(&copy[0], &grid[0], nodes * sizeof(grid[0]))) {
    fprintf(stderr, "validation failed\n");
    return EXIT_FAILURE;
  }
  t = now() - t;
  fprintf(stderr, "seconds=%.3f\n", t);
#endif

  if (buffer)
    delete[] buffer;

  fprintf(stderr, "OK\n");

  return 0;
}
