// hzip example illustrating compression of unstructured quad and hex meshes

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/time.h>
#include <vector>
#include "hzip.h"

using namespace std;

#if !defined TYPE || !defined HZTYPE
  #error "must define TYPE and HZTYPE"
#endif

static double
now()
{
  struct timeval tv;
  gettimeofday(&tv, 0);
  return tv.tv_sec + 1e-6 * tv.tv_usec;
}

static void
read(FILE* file, vector<int>& mesh, vector<TYPE> node[3], unsigned& dims, unsigned& perm)
{
  dims = 0;
  perm = 0;
  char line[0x10000];
  while (fgets(line, sizeof(line), file)) {
    double p[3];
    int v[8];
    unsigned nodes, cells;
    if (sscanf(line, "v%lf%lf%lf", p+0, p+1, p+2) == 3) {
      for (unsigned i = 0; i < 3; i++)
        node[i].push_back((TYPE)p[i]);
    }
    else if (sscanf(line, "f%i%i%i%i", v+0, v+1, v+2, v+3) == 4) {
      for (unsigned i = 0; i < 4; i++)
        mesh.push_back(v[i] > 0 ? v[i] - 1 : v[i] + node[0].size());
      dims = 2;
      perm = 0x2310;
    }
    else if (sscanf(line, "h%i%i%i%i%i%i%i%i", v+0, v+1, v+2, v+3, v+4, v+5, v+6, v+7) == 8) {
      for (unsigned i = 0; i < 8; i++)
        mesh.push_back(v[i] > 0 ? v[i] - 1 : v[i] + node[0].size());
      dims = 3;
    }
    else if (strstr(line, "format GAMBIT"))
      perm = 0x76543210;
    else if (strstr(line, "format CUBIT"))
      perm = 0x67542310;
    else if (sscanf(line, "#%u%u", &nodes, &cells) == 2) {
      for (unsigned i = 0; i < 3; i++)
        node[i].reserve(nodes);
      mesh.reserve(8 * cells);
    }
  }
}

int
main(int argc, char* argv[])
{
  char* infile = 0;
  char* meshfile = 0;
  char* nodefile = 0;
  unsigned bits = 12;

  switch (argc) {
    case 5:
      nodefile = argv[4];
      /*FALLTHROUGH*/
    case 3:
      meshfile = argv[3];
      /*FALLTHROUGH*/
    case 4:
      if (sscanf(argv[2], "%u", &bits) != 1)
        return EXIT_FAILURE;
      /*FALLTHROUGH*/
    case 2:
      infile = argv[1];
      break;
    default:
      fprintf(stderr, "Usage: %s <infile> [bits] [mesh file] [node file]\n", *argv);
      return EXIT_FAILURE;
  }

  // read uncompressed quad or hex mesh
  fprintf(stderr, "reading\n");
  FILE* file = fopen(infile, "r");
  if (!file) {
    fprintf(stderr, "cannot open file\n");
    return EXIT_FAILURE;
  }
  vector<int> mesh;
  vector<TYPE> node[3];
  unsigned dims;
  unsigned perm;
  read(file, mesh, node, dims, perm);
  fclose(file);
  fprintf(stderr, "bytes=%u\n", (unsigned)(mesh.capacity() * sizeof(int) + (node[0].capacity() + node[1].capacity() + node[2].capacity()) * sizeof(TYPE)));

  unsigned nodes = node[0].size();
  unsigned indices = mesh.size();
  unsigned cells = indices >> dims;
  unsigned char* buffer;
  double t;
  int inbytes;
  int outbytes;

  // compress/decompress mesh connectivity

  HZMstream* mstream;
  #ifndef WITHOUT_ZLIB
    #define CODEC HZM_CODEC_ZLIB
    HZMCODECzlib param = hzm_codec_zlib;
  #else
    #define CODEC HZM_CODEC_BASE
    HZMCODECbase param = hzm_codec_base;
  #endif
  param.bits = bits;
  inbytes = indices * sizeof(int);

  t = now();
  if (meshfile) {
    // compress mesh to file
    fprintf(stderr, "compressing mesh to file\n");
    file = fopen(meshfile, "wb");
    if (!file) {
      fprintf(stderr, "cannot create file\n");
      return EXIT_FAILURE;
    }
    mstream = hzip_mesh_create_file(file, perm, dims, cells, CODEC, &param);
    buffer = 0;
  }
  else {
    // compress mesh to memory
    fprintf(stderr, "compressing mesh to memory\n");
    buffer = new unsigned char[inbytes];
    mstream = hzip_mesh_create_mem(buffer, inbytes, perm, dims, cells, CODEC, &param);
    file = 0;
  }
  // execute compression
  if (!mstream ||
      hzip_mesh_write(mstream, &mesh[0], cells) < 0 ||
      (outbytes = hzip_mesh_close(mstream)) < 0) {
    fprintf(stderr, "compression failed\n");
    return EXIT_FAILURE;
  }
  t = now() - t;
  fprintf(stderr, "in=%u out=%u ratio=%.2f bpc=%.2f seconds=%.3f cps=%.0f\n", inbytes, outbytes, (double)inbytes / outbytes, outbytes * 8.0 / cells, t, cells / t);

  if (file)
    fclose(file);

#ifndef WITHOUT_DECOMPRESSION
  t = now();
  if (meshfile) {
    // decompress mesh from file
    fprintf(stderr, "decompressing mesh from file\n");
    file = fopen(meshfile, "rb");
    if (!file) {
      fprintf(stderr, "cannot open file\n");
      return EXIT_FAILURE;
    }
    mstream = hzip_mesh_open_file(file, perm);
  }
  else {
    // decompress mesh from memory
    fprintf(stderr, "decompressing mesh from memory\n");
    mstream = hzip_mesh_open_mem(buffer, inbytes, perm);
  }
  // execute decompression
  vector<int> mcopy(indices);
  if (!mstream ||
      hzip_mesh_dimensions(mstream) != dims ||
      hzip_mesh_cells(mstream) != cells ||
      hzip_mesh_read(mstream, &mcopy[0], cells) < 0) {
    fprintf(stderr, "decompression failed\n");
    return EXIT_FAILURE;
  }
  if (hzip_mesh_close(mstream) != outbytes)
    fprintf(stderr, "ERROR: unexpected byte count\n");
  t = now() - t;
  fprintf(stderr, "seconds=%.3f cps=%.0f\n", t, cells / t);

  if (file)
    fclose(file);

  // validate data
  fprintf(stderr, "validating\n");
  t = now();
  if (memcmp(&mcopy[0], &mesh[0], indices * sizeof(mesh[0]))) {
    fprintf(stderr, "validation failed\n");
    return EXIT_FAILURE;
  }
  t = now() - t;
  fprintf(stderr, "seconds=%.3f\n", t);
#endif

  if (buffer)
    delete[] buffer;

  // compress node geometry

  HZNstream* nstream;
  inbytes = 3 * nodes * sizeof(node[0][0]);
  outbytes = 0;

  t = now();
  if (nodefile) {
    // compress nodes to file
    fprintf(stderr, "compressing nodes to file\n");
    file = fopen(nodefile, "wb");
    if (!file) {
      fprintf(stderr, "cannot create file\n");
      return EXIT_FAILURE;
    }
    buffer = 0;
  }
  else {
    // compress nodes to memory
    fprintf(stderr, "compressing nodes to memory\n");
    buffer = new unsigned char[inbytes];
    file = 0;
  }
  // execute compression
  for (unsigned i = 0; i < 3; i++) {
    if (file)
      nstream = hzip_node_create_file(file, perm, 0, dims, nodes, HZTYPE, CODEC, 0);
    else
      nstream = hzip_node_create_mem(buffer + outbytes, inbytes - outbytes, perm, 0, dims, nodes, HZTYPE, CODEC, 0);
    int status;
    if (!nstream ||
        hzip_node_write(nstream, &node[i][0], &mesh[0], cells) < 0 ||
        (status = hzip_node_close(nstream)) < 0) {
      fprintf(stderr, "compression failed\n");
      return EXIT_FAILURE;
    }
    outbytes += status;
  }
  t = now() - t;
  fprintf(stderr, "in=%u out=%u ratio=%.2f bpn=%.2f seconds=%.3f nps=%.0f\n", inbytes, outbytes, (double)inbytes / outbytes, outbytes * 8.0 / nodes, t, nodes / t);

  if (file)
    fclose(file);

#ifndef WITHOUT_DECOMPRESSION
  t = now();
  if (nodefile) {
    // decompress nodes from file
    fprintf(stderr, "decompressing nodes from file\n");
    file = fopen(nodefile, "rb");
    if (!file) {
      fprintf(stderr, "cannot open file\n");
      return EXIT_FAILURE;
    }
  }
  else {
    // decompress nodes from memory
    fprintf(stderr, "decompressing nodes from memory\n");
  }
  // ececute decompression
  vector<TYPE> ncopy[3];
  int offset = 0;
  for (unsigned i = 0; i < 3; i++) {
    ncopy[i] = vector<TYPE>(nodes);
    if (file)
      nstream = hzip_node_open_file(file, perm, 0);
    else
      nstream = hzip_node_open_mem(buffer + offset, inbytes - offset, perm, 0);
    int status;
    if (!nstream ||
        hzip_node_dimensions(nstream) != dims ||
        hzip_node_count(nstream) != nodes ||
        hzip_node_type(nstream) != HZTYPE ||
        hzip_node_read(nstream, &ncopy[i][0], &mesh[0], cells) != (int)nodes ||
        (status = hzip_node_close(nstream)) < 0) {
      fprintf(stderr, "decompression failed\n");
      return EXIT_FAILURE;
    }
    offset += status;
  }
  if (offset != outbytes)
    fprintf(stderr, "ERROR: unexpected byte count\n");
  t = now() - t;
  fprintf(stderr, "seconds=%.3f nps=%.0f\n", t, nodes / t);

  if (file)
    fclose(file);

  // validate data
  fprintf(stderr, "validating\n");
  t = now();
  for (unsigned i = 0; i < 3; i++)
    if (memcmp(&ncopy[i][0], &node[i][0], nodes * sizeof(node[i][0]))) {
      fprintf(stderr, "validation failed\n");
      return EXIT_FAILURE;
    }
  t = now() - t;
  fprintf(stderr, "seconds=%.3f\n", t);
#endif

  if (buffer)
    delete[] buffer;

  fprintf(stderr, "OK\n");

  return 0;
}
